/*-
 * Copyright (C)2020 @BABOLO http://www.babolo.ru/
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#ident "@(#) Copyright (C)2020 @BABOLO http://www.babolo.ru/"
#ident "@(#) $Id: util.c,v 1.2 2020/05/16 01:01:00 babolo Exp $"

#define MIFE_COMPAT     5
#define BLIN_COMPAT     3
#define RECOBE_COMPAT   4
#define RECOBE_INTERNAL 1

#include <sys/types.h>
#include <sys/event.h>
#include <sys/time.h>
#include <sysexits.h>
#include <string.h>
#include <stdlib.h>
#include <errno.h>
#include <stdio.h>
#include <babolo/BLINflag.h>
#include <mife.h>
#include "recobe.h"

void
recobe_cnverr(const char *file, const char *func, int line, const char *mesg, recobe_conv cnv) {
    const recobe_c_tnm *encolist = recobe_collection.domains[cnv.domain]->encolist;

    blin_cry1( 0
             , "%s:%s:%d %s %s=%c%c(%d: %c%d -> %c%d)"
             , file
             , func
             , line
             , mesg
             , recobe_collection.domains[cnv.domain]
             ? recobe_collection.domains[cnv.domain]->domname
             : "-"
             , cnv.nonesrc ? '.' : encolist[cnv.encosrc].encoid ? encolist[cnv.encosrc].encoid[0] : '-'
             , cnv.nonedst ? '.' : encolist[cnv.encodst].encoid ? encolist[cnv.encodst].encoid[0] : '-'
             , cnv.domain
             , cnv.nonesrc ? '-' : '+'
             , cnv.encosrc
             , cnv.nonedst ? '-' : '+'
             , cnv.encodst
             );
}

void
recobe_testtail(const char *file, const char *func, int line, recobe_chunk *chunk) {
    if  (!!chunk) {
        blin_cry1( 3
                 , "%s:%s:%d chunk=%"BLIN_X" %u[%u]\n"
                 , file
                 , func
                 , line
                 , BLIN_I(chunk)
                 , chunk->position
                 , chunk->length
                 )
        ;
        fprintf(stderr, "");
        for (u_int32_t i = chunk->position; i < chunk->length; ++i) {
            if  ((chunk->chunk[i] < ' ') || ((chunk->chunk[i] >= 0x7F) && (chunk->chunk[i] < 0xC0))) {
                fprintf(stderr, "\\%02X", chunk->chunk[i]);
            } else {
                fprintf(stderr, "%c", chunk->chunk[i]);
        }   }
        fprintf(stderr, "\n");
}   }

void
recobe_testchain(const char *file, const char *func, int line, recobe_chain *chain) {
    if  (!!chain) {
        const recobe_c_tnm *encolist = recobe_collection.domains[chain->cnv.domain]->encolist;

        blin_cry1( 3
                 , "%s:%s:%d chain=%"BLIN_X" f%08X cnv %s=%c%c(%d: %c%d=>%c%d) %ux%u=>%ux%u ~%s~%s~\n"
                   " next=%"BLIN_X" convertor=%"BLIN_X" converdata=%"BLIN_X" tail=%"BLIN_X"\n"
                 , file
                 , func
                 , line
                 , BLIN_I(chain)
                 , chain->flags
                 , recobe_collection.domains[chain->cnv.domain]->domname
                 , chain->cnv.nonesrc ? '.' : encolist[chain->cnv.encosrc].encoid[0]
                 , chain->cnv.nonedst ? '.' : encolist[chain->cnv.encodst].encoid[0]
                 , chain->cnv.domain
                 , chain->cnv.nonesrc ? '-' : '+'
                 , chain->cnv.encosrc
                 , chain->cnv.nonedst ? '-' : '+'
                 , chain->cnv.encodst
                 , chain->inwlen
                 , chain->inwnum
                 , chain->outwlen
                 , chain->outwnum
                 , chain->bsubst ? chain->bsubst : (u_char*)""
                 , chain->esubst ? chain->esubst : (u_char*)""
                 , BLIN_I(chain->next)
                 , BLIN_I(chain->convertor)
                 , BLIN_I(chain->converdata)
                 , BLIN_I(chain->tail)
                 )
        ;
        if  (chain->tail) recobe_testtail(file, func, line, chain->tail);
        if  (chain->next) recobe_testchain(file, func, line, chain->next);
}   }

u_int32_t
recobe_chunksize(recobe_chunk *chunk) {
    u_int32_t size = 0;

    if  (!!chunk && (chunk->length > chunk->position)) {
        size = chunk->length - chunk->position;
    }
    return(size);
}

const u_char *
recobe_chunktext(recobe_chunk *chunk) {
    const u_char *txt = NULL;

    if  (!!chunk) txt = &chunk->chunk[chunk->position];
    return(txt);
}

recobe_chunk *
recobe_chunkeot(recobe_chain *chain) {
    recobe_chunk *eot;

#   define blin_internal_flags (chain->flags)
    if  (!(eot = calloc(1, sizeof(recobe_chunk)))) {
        ifBLIN_QW0("calloc(%"BLIN_D")", sizeof(recobe_chunk));
    } else {
        eot->position = RECOBE_EOT;
    }
    return(eot);
#   undef blin_internal_flags
}

recobe_chunk *
recobe_chunkopy(recobe_chain *chain) {
    recobe_chunk *kopy;

#   define blin_internal_flags (chain->flags)
    if  (!(kopy = malloc(sizeof(recobe_chunk) + recobe_chunksize(chain->tail)))) {
        ifBLIN_QW0("malloc(%"BLIN_D")", sizeof(recobe_chunk) + recobe_chunksize(chain->tail));
    }
    bcopy(chain->tail, kopy, sizeof(recobe_chunk) + recobe_chunksize(chain->tail));
    return(kopy);
#   undef blin_internal_flags
}

/* recobe_chunksum() < 0 - 
 *                   == 1 -  source
 *                   == 2 -  
 *                   == 0 - ...
 */
int
recobe_chunksum(recobe_chain *chain, recobe_chunk *source) {
    u_int32_t     tailen = 0;
    u_int32_t     srclen = 0;
    int           ex = EX_OK;
    recobe_chunk *sum;

#   define blin_internal_flags (chain->flags)
    if  (!source) {
        ifBLIN_QW0("source");
        errno = EFAULT;
        ex = -EX_SOFTWARE;
        goto out;
    }
    tailen = recobe_chunksize(chain->tail);
    if  (!(srclen = recobe_chunksize(source))) {
        if  (RECOBE_EOT == source->position) {
            ex = 1;
            if  (!tailen) ex = 2;
        }
        goto out;
    }
    if  (!(sum = malloc(sizeof(recobe_chunk) + tailen + srclen))) {
        ifBLIN_QW0("malloc(%"BLIN_D")", sizeof(recobe_chunk) + tailen + srclen);
        goto out;
    }
    sum->position = 0;
    sum->length = tailen + srclen;
    if  (!!tailen) {
        memcpy(sum->chunk, recobe_chunktext(chain->tail), tailen);
    }
    if  (!!chain->tail) free(chain->tail);
    chain->tail = NULL;
    if  (!!srclen) {
        memcpy(&sum->chunk[tailen], recobe_chunktext(source), srclen);
    }
    chain->tail = sum;
out:
    return(ex);
#   undef blin_internal_flags
}

static int
convertor(recobe_chain *chain, recobe_chunk *chunk) {
    u_char       *table = (u_char*)chain->converdata;
    int           ex;

#   define blin_internal_flags (chain->flags)
    ifBLIN_QX3("- %08X", chain->flags);
    chunk->chunk[chunk->length] = table[*(recobe_chunktext(chain->tail))];
    if  (  !!chain->tail->chunk[chain->tail->position++]
        && !chunk->chunk[chunk->length]
        ) {
        ex = -1;
        goto out;
    }
    chunk->length++;
    ex = 1;
out:
    ifBLIN_QX3("- %d", ex);
    return(ex);
#   undef blin_internal_flags
}

recobe_chain *
recobe_openc(recobe_conf *cf, const char *p, int domain, recobe_chain *next) {
#   define blin_internal_flags (cf->flags)
    recobe_chain *chain = NULL;

    ifBLIN_QX3("+ d%d:%s", domain, p);
    if  (!(chain = malloc(sizeof(recobe_chain)))) {
        ifBLIN_QW0("malloc(%d)", sizeof(recobe_chain));
        goto out;
    }
    chain->cf = cf;
    chain->flags = cf->flags & ~(RECOBE_USED | RECOBE_CHAINER);
    chain->tail = NULL;
    chain->next = next;
    chain->cnv = recobe_collection.cnvby(p, domain);
    chain->convertor = convertor;
    if  (  !recobe_collection.domains[chain->cnv.domain]->reconv
        || !(chain->converdata = recobe_collection.domains[chain->cnv.domain]->reconv(chain->cnv))
        ) {
        ifBLIN_QX0("Illegal reconv -c %s", p);
        free(chain);
        chain = NULL;
        goto out;
    }
    chain->bsubst = cf->bsubst;
    chain->esubst = cf->esubst;
    chain->inwlen = 1;
    chain->outwlen = 1;
    chain->inwnum = 1;
    chain->outwnum = 1;
out:
    RECOBE_CHAIN(chain);
    ifBLIN_QX3("- %"BLIN_X, chain);
    return(chain);
#   undef blin_internal_flags
}
